/* Copyright (c) 2000-2006 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */


const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

var myPrefWindow = null;
var startBox;

var gInvocation = null;
var gInvokeType = 0; // what kind of invocation are we displaying?

var gInvocationPrefs = null;
var gProjectInvocationPrefs = null;
var gDebugPrefs = null;
var gMainWindow= null;

var dialog = document.getElementById('dialog-debugging-properties');
dialog.log = ko.logging.getLoggingMgr().getLogger("debugging-properties");

function SetupDialog() {
    try {
        gMainWindow = ko.windowManager.getMainWindow();
        myPrefWindow = new koPrefWindow(null);

        gInvocation = window.arguments[0].invocation;
        gInvokeType = window.arguments[0].invoke_type;
        gDebugPrefs = window.arguments[0].prefs;
        gInvocationPrefs = Components.classes["@activestate.com/koPreferenceSet;1"].createInstance();
        gInvocationPrefs.id = "Invocations";
        if (gDebugPrefs.parent.hasPref("Invocations")) {
            gProjectInvocationPrefs = gDebugPrefs.parent.getPref("Invocations");
        }        
        if (gDebugPrefs.hasPrefHere("Invocations")) {
            gInvocationPrefs.update(gDebugPrefs.getPref('Invocations'));
        } else if (!gProjectInvocationPrefs) {
            // no preexisting invocation prefs, set the ones we have now
            gInvocationPrefs.setPref(gInvocation.persistableInstancePreferences.id,
                                     gInvocation.persistableInstancePreferences);
        }
        
        UpdateLabelsForInvokeType();
        FillInvocationPopup();
        UpdateInvocation();

        gMainWindow.gOpeningDialog = false;
    } catch(e) {
        dialog.log.exception(e);
    }
}

function UpdateLabelsForInvokeType() {
    try {
        var bundle = Components.classes["@mozilla.org/intl/stringbundle;1"]
                .getService(Components.interfaces.nsIStringBundleService)
                .createBundle("chrome://komodo/locale/debugger.properties");

        switch (gInvokeType) {
            case Components.interfaces.koIInvocation.INVOKE_RUN:
                document.title = bundle.GetStringFromName("invokeRun.title");
                break;
            case Components.interfaces.koIInvocation.INVOKE_PROFILE:
                document.title = bundle.GetStringFromName("invokeProfile.title");
                break;
            default:
                // Just leave it as it already is.
                break;
        }
    } catch(e) {
        dialog.log.exception(e);
    }
}

function UpdateInvocation() {
    try {

        var prefs = gInvocation.currentInstancePreferences;
        var language = null;
        if (prefs.hasStringPref("language"))
            language = prefs.getStringPref("language");
        if (!language)
            language = gInvocation.name;

        if (language) {
            document.getElementById('language').value = language;
            // add the debugger component to the ui
            var interpparams = document.getElementById('interpparams');
            if (interpparams.lastChild.nodeName == 'debugOptions')
                interpparams.removeChild(interpparams.lastChild)
            var el = document.createElementNS(XUL_NS, 'debugOptions');
            el.setAttribute('type', language.toLowerCase())
            interpparams.appendChild(el);

            var propPanel = document.getElementById('generalProperties');
            if (propPanel.lastChild.nodeName == 'debugProperties')
                propPanel.removeChild(propPanel.lastChild)
            el = document.createElementNS(XUL_NS, 'debugProperties');
            el.setAttribute('type', language.toLowerCase())
            el.setAttribute('prefwidgetauto', 'true');
            propPanel.appendChild(el);
            try {
                el.init();
            } catch(e) {
                // do nothing
            }
        }

        myPrefWindow.init(gInvocation.currentInstancePreferences);
        myPrefWindow.onpageload();
        toggleCGITabs();
        // for convenience, focus the params field.  "tabindex=" will make this redundant when it works!
        document.getElementById("params").focus();
    } catch(e) {
        dialog.log.exception(e);
    }
}

function FillInvocationPopup() {
    try {
        var menupopup = document.getElementById('invocation-configuration-popup');
        while (menupopup.childNodes.length) {
            menupopup.removeChild(menupopup.childNodes[0]);
        }
        var configurations = new Object();
        gInvocationPrefs.getPrefIds(configurations, new Object());
        //dump('got '+configurations.value.length+' invocations for this file\n');
        if (configurations.value.length <= 1)
            document.getElementById('deleteButton').setAttribute('disabled','true');
        configurations = configurations.value;
        
        if (gProjectInvocationPrefs) {
            var projConfig = new Object();
            gProjectInvocationPrefs.getPrefIds(projConfig, new Object());
            configurations = configurations.concat(projConfig.value);
        }
        //dump("current invocation name "+currentInvocation.currentInstancePreferences.id+"\n");
        //currentInvocation.currentInstancePreferences.dump(4);
        //gInvocationPrefs.dump(0);
        for (var i=0; i < configurations.length; i++) {
            var invocation_pref = null;
            if (gInvocationPrefs.hasPrefHere(configurations[i])) {
                invocation_pref = gInvocationPrefs.getPref(configurations[i]);
            } else if (gProjectInvocationPrefs &&
                       gProjectInvocationPrefs.hasPref(configurations[i])) {
                invocation_pref = gProjectInvocationPrefs.getPref(configurations[i]);
            }
            var menuitem = document.createElementNS(XUL_NS, 'menuitem');
            menuitem.setAttribute('label', configurations[i]);
            menupopup.appendChild(menuitem);
            //dump("Adding config "+configurations[i]+"\n");
            //dump("       config "+invocation_pref.id+"\n");
            //invocation_pref.dump(4)
            if (invocation_pref.id == gInvocation.currentInstancePreferences.id) {
                document.getElementById('invocation-configuration').setAttribute('label', configurations[i]);
            }
        }
    } catch(e) {
        dialog.log.exception(e);
    }
}

function newPrefset()
{
    // save any settings in the current prefs
    var oldPrefs = gInvocation.currentInstancePreferences;

    // create a new default set
    gInvocation = Components.classes["@activestate.com/koInvocation;1?type="+gInvocation.name].createInstance();
    var catOb = new Object();
    var instOb = new Object();
    gInvocation.getDefaultPreferences(catOb, instOb);
    gInvocation.currentCategoryPreferences = catOb.value;
    gInvocation.currentInstancePreferences = instOb.value;
    gInvocation.currentInstancePreferences.update(oldPrefs);
}

function setCurrentPrefs(prefs)
{
    // create a new default set, but use the prefs
    // passed in as the new current instance prefs
    var catOb = new Object();
    var instOb = new Object();
    var langName = gInvocation.name;
    //if (prefs.hasStringPref("language")) {
    //    var new_langName = prefs.getStringPref("language");
    //    if (langName && !new_langName) {
    //        dump("debugging-properties.js :: setCurrentPrefs :: "
    //             + ", gInvocation.name:"
    //             + gInvocation.name
    //             + ', but prefs.getStringPref("language") => null\n');
    //    }
    //}
    if (langName != gInvocation.name) {
        gInvocation = Components.classes["@activestate.com/koInvocation;1?type="+langName].createInstance();
        //dump("debugging-properties.js :: setCurrentPrefs :: langName:"
        //     + langName
        //     + ", gInvocation.name:"
        //     + gInvocation.name
        //     + "\n");
    }
    gInvocation.getDefaultPreferences(catOb, instOb);
    gInvocation.currentCategoryPreferences = catOb.value;
    gInvocation.currentInstancePreferences = instOb.value;
    gInvocation.currentInstancePreferences.update(prefs);
    gInvocation.currentInstancePreferences.id = prefs.id
    gInvocation.currentInstancePreferences.setStringPref("language", langName);
}

function ChangeInvocation() {
    // be sure the set we're leaving is saved
    if (gInvocationPrefs.hasPref(gInvocation.persistableInstancePreferences.id))
        gInvocationPrefs.setPref(gInvocation.persistableInstancePreferences.id,
                             gInvocation.persistableInstancePreferences);

    var name = document.getElementById('invocation-configuration').getAttribute("label");
    var invocation_pref = null;

    if (gInvocationPrefs.hasPrefHere(name)) {
        invocation_pref = gInvocationPrefs.getPref(name);
    } else if (gProjectInvocationPrefs &&
               gProjectInvocationPrefs.hasPref(name)) {
        invocation_pref = gProjectInvocationPrefs.getPref(name);
    }

    invocation_pref.id = name;
    setCurrentPrefs(invocation_pref);
    UpdateInvocation();
}

function doSaveNewInvocation() {
    // first, get a name for the new invocation
    try {
        var name = ko.dialogs.prompt("Enter a new configuration name.  The new "+
                                 "configuration will be based on the currently "+
                                 "selected '"+gInvocation.currentInstancePreferences.id+"' configuration.",
                                 "Configuration Name", "",
                                 "Configuration Name");
        if (!name) return;
        if (gInvocationPrefs.hasPref(name) || (gProjectInvocationPrefs &&
               gProjectInvocationPrefs.hasPref(name))) {
            ko.dialogs.alert("A debugging configuration with the name '"+name+"' already exists.\n");
            return;
        }
        // be sure the set we're leaving is saved
        if (gInvocationPrefs.hasPref(gInvocation.persistableInstancePreferences.id))
            gInvocationPrefs.setPref(gInvocation.persistableInstancePreferences.id,
                                 gInvocation.persistableInstancePreferences);

        // create a new debug data set
        newPrefset();
        gInvocation.currentInstancePreferences.id = name;
        gInvocationPrefs.setPref(gInvocation.persistableInstancePreferences.id,
                                 gInvocation.persistableInstancePreferences);

        //gInvocation.currentInstancePreferences.dump(0)

        FillInvocationPopup();
        UpdateInvocation();
        document.getElementById('deleteButton').removeAttribute('disabled');
    } catch(e) {
        dialog.log.exception(e);
    }
}

function doDeleteInvocation() {
    try {
        var name = document.getElementById('invocation-configuration').getAttribute('label');
        if (!gInvocationPrefs.hasPrefHere(name)) {
            return;
        } else {
            if (ko.dialogs.yesNo("Are you sure you want to delete the "+name+" configuration?") != "Yes") return;
        }
        gInvocationPrefs.deletePref(name);

        var configurations = new Object();
        gInvocationPrefs.getPrefIds(configurations, new Object());
        var invocation_prefs = gInvocationPrefs.getPref(configurations.value[0]);
        setCurrentPrefs(invocation_prefs);

        FillInvocationPopup();
        UpdateInvocation();
    } catch(e) {
        dialog.log.exception(e);
    }
}

function _setReturnVar(name, val) {
  if (window.arguments && window.arguments[0])
    window.arguments[0][name]=val;
  return true;
}

function _setResult(res) {
    return _setReturnVar("res", res)
}

function Ok() {
    try {
        // get the invocation to validate the preference set.
        if (!myPrefWindow._onOK()) {
            UpdateInvocation();
            return false;
        }
        var bad_pref_id = new Object();
        var message = new Object();
        var ok = gInvocation.validate(gInvokeType, bad_pref_id, message)
        if (!ok) {
            var badWidget = myPrefWindow.findElementWithPreferenceId(bad_pref_id.value);
            if (badWidget)
                badWidget.focus();
            ////XXX Refactor: Copy of dialog invocation sequence in debugger.js::_showInvalidInterpreterMessageBox
            var result = ko.dialogs.customButtons(message.value,
                                              ['OK','&Help','&Preferences'],
                                              'OK', null, 'Debugger Error');
            if (result == 'Help') {
                ko.help.open(gInvocation.name.toLowerCase() + "_configuration");
            } else if (result == 'Preferences') {
                ko.windowManager.getMainWindow().
                    prefs_doGlobalPrefs(gInvocation.name.toLowerCase() + "Item");
            }
            UpdateInvocation();
            return false;
        }

        // save our prefs back into the files prefs now
        if (!gDebugPrefs.hasPref("Invocations")) {
            gDebugPrefs.setPref('Invocations', gInvocationPrefs);
        } else {
            // first, delete any prefs that are in gDebugPrefs but not in
            // gInvocationPrefs
            var debugInvocations = gDebugPrefs.getPref('Invocations');
            var newConfig = new Object();
            gInvocationPrefs.getPrefIds(newConfig, new Object());
            newConfig = newConfig.value;
            var oldConfig = new Object();
            debugInvocations.getPrefIds(oldConfig, new Object());
            oldConfig = oldConfig.value;
            for (var c=0; c < oldConfig.length; c++) {
                if (newConfig.indexOf(oldConfig[c]) < 0) {
                    debugInvocations.deletePref(oldConfig[c]);
                }
            }
            debugInvocations.update(gInvocationPrefs);
        }
        window.arguments[0].invocation=gInvocation;
        window.arguments[0].res="ok";
    } catch(e) {
        dialog.log.exception(e);
    }
    return true;
}

function Cancel() {
    window.arguments[0].invocation=null;
    window.arguments[0].res="cancel";
    return true;
}

function doClose() {
    Cancel();
    window.close();
}

// Utility functions for the various panels.
function doBrowseForDir(widget_id) {
    if (!widget_id)
        widget_id = "cwd";
    var textField = document.getElementById(widget_id);
    var dir = ko.filepicker.getFolder(textField.value);
    if (dir) textField.value = dir;
}

function doBrowseForLanguageFile(widget_id, language) {
    if (typeof(language) == 'undefined') language = gInvocation.name;
    var textField = document.getElementById(widget_id);
    var path = ko.filepicker.browseForFile(null, textField.value,
                                   null, // title
                                   language); // default filter name
    if (path != null) {
        textField.value = path;
    }
}

function doBrowseForXMLInputFile(widget_id) {
    if (!widget_id) return;
    var textField = document.getElementById(widget_id)
    var path = ko.filepicker.browseForFile(null, textField.value,
                                   null, // title
                                   "XML", // default filter name
                                   ["XML", "XSLT", "HTML", "All"]);
    if (path != null) {
        textField.value = path;
    }
}

function doBrowseForCommandFile(widget_id) {
    if (!widget_id) return;
    var textField = document.getElementById(widget_id)
    var path = ko.filepicker.browseForExeFile(null, textField.value);
    if (path != null) {
        textField.value = path;
    }
}


function toggleCGITabs() {
    try {
        var usecgi = document.getElementById("debuggerRunAsCGI");
        if (!usecgi) return;
        var checked = usecgi.getAttribute("checked");
        var t1 = document.getElementById("tab-cgiEnvironment");
        var t2 = document.getElementById("tab-cgiInput");
        if (checked=="true") {
            t1.removeAttribute('hidden');
            t2.removeAttribute('hidden');
        } else {
            t1.setAttribute('hidden','true');
            t2.setAttribute('hidden','true');
        }
    } catch (e) {
        dialog.log.exception(e);
    }
}

function updateInterpreterArgumentsWithDir(menuitem) {
    try {
        var dir = ko.filepicker.getFolder();
        if (!dir) return;
        var val = menuitem.getAttribute('value');
        var params = document.getElementById('executable-params');
        if (params.value) {
            params.value += ' '+val+' '+dir;
        } else {
            params.value = val+' '+dir;
        }
        document.getElementById('executable-params').focus();
    } catch (e) {
        dialog.log.exception(e);
    }
}

function updateInterpreterArguments(menuitem) {
    try {
        var val = menuitem.getAttribute('value');
        var params = document.getElementById('executable-params');
        if (params.value) {
            params.value += ' '+val;
        } else {
            params.value = val;
        }
        document.getElementById('executable-params').focus();
    } catch (e) {
        dialog.log.exception(e);
    }
}
